//! This is a skeleton Radicle CI adapter for demonstration purposes.
//!
//! This program shows how to build a very simple adapter for Radicle
//! CI. As per the CI broker/adapter protocol, it reads a trigger
//! request message from its stdin, and writes two response messages
//! to its stdout. Based on the trigger, it clones the relevant Git
//! repository, and checks out the relevant Git commit, then runs
//! executed a silly CI plan. To keep things simple, and quick, in
//! this program the plan is "does the .radicle directory exist?". If
//! it exists, CI succeeds, otherwise it fails. If any of the previous
//! steps fails, CI also fails.
//!
//! Any errors are reported to stderr. If things go wrong very early,
//! the first response message may be a failure message, rather than a
//! "triggered" message. The CI broke will hopefully handle this.

use std::{error::Error, fs::create_dir_all, path::Path, time::SystemTime};

use clap::Parser;
use html_page::{Element, HtmlPage, Tag};
use tempfile::tempdir;

use radicle_ci_broker::msg::{
    helper::{get_sources, read_request, write_failed, write_succeeded, write_triggered, AdminLog},
    Request, RunId, RunResult,
};

use radicle_ci_ambient::{
    ambient::{self, run_ambient_ci},
    config::Config,
    AdapterError,
};

// Name of environment variable that has filename for configuration
// file.
const CONFIG_ENV: &str = "RADICLE_CI_AMBIENT";

fn main() -> Result<(), AdapterError> {
    match fallible_main() {
        Ok(true) => write_succeeded().map_err(AdapterError::Succeded)?,
        Ok(false) => write_failed().map_err(AdapterError::Failed)?,
        Err(err) => {
            eprintln!("ERROR: {err}");
            write_failed().map_err(AdapterError::Failed)?;
            std::process::exit(1);
        }
    }

    Ok(())
}

fn fallible_main() -> Result<bool, AdapterError> {
    let args = Args::parse();

    let config = Config::load_via_env(CONFIG_ENV)?;

    if args.show_config {
        eprintln!("{config:#?}");
        return Ok(false);
    }

    let mut adminlog = config.open_log()?;
    adminlog.writeln("radicle-ci-ambient starts")?;

    // Read the trigger request message and extract the repository ID
    // and commit from it, if possible.
    let req = read_request().map_err(AdapterError::ReadRequest)?;
    let repoid = req.repo();
    let commit = req.commit().map_err(AdapterError::Commit)?;

    // Create a temporary directory and a sub-directory `src` where
    // we'll put the source code from Git.
    let tmp = tempdir().map_err(AdapterError::TempDir)?;
    let src = tmp.path().join("src");
    get_sources(&mut adminlog, args.dry_run, repoid, commit, &src)?;

    // Info about this CI run.
    let mut info = RunInfoBuilder::default();

    // Invent an adapter run ID and tell it to the CI broker.
    let run_id = RunId::default();
    info.set_adapter_run_id(&run_id);
    let project_name = repoid.to_string();
    let project_log_dir = config.logdir.join(&project_name);
    if !project_log_dir.exists() {
        create_dir_all(&project_log_dir)
            .map_err(|err| AdapterError::CreateDirs(project_log_dir, err))?;
    }
    let run_log_relative = format!("{project_name}/{run_id}.html");

    let url = config
        .base_url
        .clone()
        .map(|base| format!("{base}/{run_log_relative}"));
    write_triggered(&run_id, url.as_deref()).map_err(AdapterError::Triggered)?;

    // Execute CI in the source directory.
    let result = execute_ci(
        &mut adminlog,
        args.dry_run,
        &config,
        info,
        &req,
        &project_name,
        &run_log_relative,
        &src,
    );
    adminlog.writeln(&format!(
        "radicle-ci-ambient ends: Ambient result is {result:?}"
    ))?;

    if let Err(err) = &result {
        adminlog.writeln(&format!("ERROR: {err}"))?;
        let mut err = err.source();
        while let Some(cause) = err {
            adminlog.writeln(&format!("caused by {cause}"))?;
            err = cause.source();
        }
    }

    result
}

#[derive(Parser)]
#[clap(version)]
struct Args {
    #[clap(long)]
    dry_run: bool,

    #[clap(long)]
    show_config: bool,
}

// This is what executes the steps CI is meant to execute. Edit this
// according to your needs.
#[allow(clippy::too_many_arguments)]
fn execute_ci(
    adminlog: &mut AdminLog,
    dry_run: bool,
    config: &Config,
    mut info: RunInfoBuilder,
    trigger: &Request,
    project_name: &str,
    run_log_relative: &str,
    src: &Path,
) -> Result<bool, AdapterError> {
    adminlog.writeln(&format!("execute_ci: src={}", src.display()))?;

    let mut distilled = if dry_run {
        ambient::DistilledPlan::default()
    } else {
        let filename = src.join(".radicle/ambient.yaml");
        ambient::DistilledPlan::read(&filename)?
    };

    if !config.allow_post_plan() {
        distilled.drop_post_plan();
    }

    let projects = ambient::ProjectsBuilder::default()
        .name(project_name)
        .source(src)
        .image(&config.image)
        .artifact_max_size(1024 * 1024 * 1024)
        .cache_max_size(20 * 1024 * 1024 * 1024)
        .plan(&distilled)
        .build();

    let (exit, maybe_run_log) = run_ambient_ci(adminlog, dry_run, &projects)?;
    if exit == 0 {
        info.set_result(RunResult::Success);
    } else {
        info.set_result(RunResult::Failure);
    }

    if let Some(run_log_raw) = maybe_run_log {
        let info = info.build()?;
        write_log_files(
            &info,
            config,
            adminlog,
            trigger,
            project_name,
            run_log_relative,
            &run_log_raw,
        )?;
    }

    if dry_run {
        adminlog.writeln("execute_ci: Ambient is pretending to be OK")?;
        Ok(true)
    } else if exit == 0 {
        adminlog.writeln("execute_ci: Ambient is OK")?;
        Ok(true)
    } else {
        adminlog.writeln(&format!("ERROR: ambient exit code {exit}"))?;
        Ok(false)
    }
}

#[derive(Default)]
struct RunInfoBuilder {
    adapter_run_id: Option<RunId>,
    result: Option<RunResult>,
}

impl RunInfoBuilder {
    fn build(self) -> Result<RunInfo, AdapterError> {
        Ok(RunInfo {
            started: SystemTime::now(),
            adapter_run_id: self
                .adapter_run_id
                .ok_or(AdapterError::Missing("adapter_run_id"))?,
            result: self.result.ok_or(AdapterError::Missing("result"))?,
        })
    }

    fn set_adapter_run_id(&mut self, id: &RunId) {
        self.adapter_run_id = Some(id.clone());
    }

    fn set_result(&mut self, result: RunResult) {
        self.result = Some(result);
    }
}

struct RunInfo {
    started: SystemTime,
    adapter_run_id: RunId,
    result: RunResult,
}

impl RunInfo {
    fn to_html(&self) -> Result<Element, AdapterError> {
        let duration = self
            .started
            .elapsed()
            .map_err(AdapterError::Elapsed)?
            .as_secs();

        Ok(Element::new(Tag::Ul)
            .with_child(
                Element::new(Tag::Li)
                    .with_text("Adapter run ID: ")
                    .with_child(
                        Element::new(Tag::Span).with_text(&self.adapter_run_id.to_string()),
                    ),
            )
            .with_child(
                Element::new(Tag::Li)
                    .with_text("Result: ")
                    .with_text(match &self.result {
                        RunResult::Success => "SUCCESS",
                        RunResult::Failure => "FAILURE",
                        _ => "Unknown",
                    }),
            )
            .with_child(Element::new(Tag::Li).with_text(&format!("Duration: {duration} seconds"))))
    }
}

fn write_log_files(
    info: &RunInfo,
    config: &Config,
    adminlog: &mut AdminLog,
    trigger: &Request,
    project_name: &str,
    run_log_relative: &str,
    run_log_raw: &[u8],
) -> Result<(), AdapterError> {
    if !config.logdir.exists() {
        adminlog.writeln("create log directory")?;
        std::fs::create_dir_all(&config.logdir)
            .map_err(|err| AdapterError::CreateDirs(config.logdir.clone(), err))?;
    }

    adminlog.writeln("write run log")?;
    let filename = config.logdir.join(run_log_relative).with_extension("bin");
    std::fs::write(&filename, run_log_raw)
        .map_err(|err| AdapterError::WriteRunLog(filename.clone(), err))?;

    adminlog.writeln("convert terminal output to HTML and write that")?;
    let page = html_log(trigger, project_name, info, run_log_raw)?;
    let filename = config.logdir.join(run_log_relative).with_extension("html");
    std::fs::write(&filename, page.to_string())
        .map_err(|err| AdapterError::WriteRunLog(filename.clone(), err))?;
    Ok(())
}

fn html_log(
    trigger: &Request,
    project_name: &str,
    info: &RunInfo,
    run_log: &[u8],
) -> Result<HtmlPage, AdapterError> {
    let title = format!("CI run for {project_name}");
    let trigger = serde_json::to_string_pretty(trigger).map_err(AdapterError::ToJson)?;
    let run_log = String::from_utf8_lossy(run_log);

    Ok(HtmlPage::default()
        .with_head_element(Element::new(Tag::Title).with_text(&title))
        .with_body_element(Element::new(Tag::H1).with_text(&title))
        .with_body_element(info.to_html()?)
        .with_body_element(Element::new(Tag::H2).with_text("Triggered by"))
        .with_body_element(Element::new(Tag::Pre).with_text(&trigger))
        .with_body_element(Element::new(Tag::H2).with_text("Run log"))
        .with_body_element(Element::new(Tag::Pre).with_text(&run_log)))
}
